import dtcli.building
import dtcli.constants
import dtcli.server_api
import typer
from typing_extensions import Annotated
import pandas as pd
from dynatrace import Dynatrace
from dynatrace.environment_v2.extensions import MonitoringConfigurationDto
from dynatrace.http_client import TOO_MANY_REQUESTS_WAIT
from rich.progress import track
from rich import print
import yaml
import dtcli
from typing import Optional, List, Dict
import json
import logging
from pathlib import Path

class MyDumper(yaml.SafeDumper):
    def increase_indent(self, flow=False, *args, **kwargs):
        return super().increase_indent(flow=flow, indentless=False)

from .models import CustomDBQuery
from .db_conversion import (
    EF2SqlExtension,
    queries_from_ef1_config,
)
from .utils import (
    build_dt_custom_device_id,
    build_dt_group_id,
    dt_murmur3,
)

logger = logging.getLogger()
logger.setLevel(logging.INFO)

EF1_EXTENSION_ID = "custom.remote.python.dbquery"
TIMEOUT = 120
TIMEFRAME = "now-6M"

app = typer.Typer()


@app.command(help="Pull EF1 db queries configurations into a spreadsheet.")
def pull(
    dt_url: Annotated[str, typer.Option(envvar="DT_URL")],
    dt_token: Annotated[str, typer.Option(envvar="DT_TOKEN")],
    output_file: Optional[str] = None or f"{EF1_EXTENSION_ID}-export.xlsx",
):
    dt = Dynatrace(
        dt_url,
        dt_token,
        too_many_requests_strategy=TOO_MANY_REQUESTS_WAIT,
        retries=3,
        log=logger,
        timeout=TIMEOUT,
    )
    configs = list(dt.extensions.list_instances(extension_id=EF1_EXTENSION_ID))
    all_queries: List[CustomDBQuery] = []
    full_configs = []

    for config in track(configs, description="Pulling EF1 configs"):
        config = config.get_full_configuration(EF1_EXTENSION_ID)
        full_config = config.json()
        properties = full_config.get("properties", {})

        group_name = properties.get("group_name") or "Custom DB Queries Group"
        device_name = properties.get("custom_device_name") or "Custom DB Queries Device"
        group_id = dt_murmur3(build_dt_group_id(group_name))
        device_id = f"CUSTOM_DEVICE-{dt_murmur3(build_dt_custom_device_id(group_id, device_name))}"

        full_config.update(
            {
                "device_id": device_id,
                "database_type": properties["database_type"],
                "group_name": properties["group_name"],
                "database_host": properties["database_host"],
                "database_name": properties["database_name"],
                "database_username": properties["database_username"],
                "custom_device_name": properties["custom_device_name"],
            }
        )

        configured_queries = queries_from_ef1_config(full_config["properties"])
        all_queries.extend(configured_queries)
        full_config.update({"queries": len(configured_queries)})
        full_config["properties"] = json.dumps(properties)
        full_configs.append(full_config)

    print("Finished pulling configs...")
    print("Adding data to document...")
    writer = pd.ExcelWriter(
        output_file,
        engine="xlsxwriter",
    )
    df = pd.DataFrame(full_configs)
    df.to_excel(writer, "Endpoints", index=False, header=True)

    df_queries = pd.DataFrame({"queries": [query.value for query in all_queries]})
    df_queries.to_excel(writer, "Queries", index=False, header=True)

    print("Closing document...")
    writer.close()
    print(f"Exported configurations available in '{output_file}'")


@app.command(help="Build EF2 SQL extensions from DB queries configuration export")
def build(
    dt_url: Annotated[str, typer.Option(envvar="DT_URL")],
    dt_token: Annotated[str, typer.Option(envvar="DT_TOKEN")],
    cert_file_path: Annotated[str, typer.Option()],
    private_key_path: Annotated[str, typer.Option()],
    input_file: Optional[str] = None or f"{EF1_EXTENSION_ID}-export.xlsx",
    directory: Optional[str] = "migrated_extensions",
    merge_endpoints: Optional[bool] = False,
    credential_vault_id: Optional[str] = None,
    upload: Optional[bool] = False,
    create_config: Optional[bool] = False,
    pre_cron: Optional[bool] = False
):

    dt = Dynatrace(
        dt_url,
        dt_token,
        too_many_requests_strategy=TOO_MANY_REQUESTS_WAIT,
        retries=3,
        log=logger,
        timeout=TIMEOUT,
    )

    xls = pd.ExcelFile(input_file)
    df = pd.read_excel(xls, "Endpoints")

    if merge_endpoints:

        grouped_endpoints: Dict[str, List] = {}

        for index, row in df.iterrows():
            config_props = json.loads(row["properties"], strict=False)
            host = config_props.get("database_host")
            database = config_props.get("database_name")
            jdbc_string = config_props.get("jdbc_connection_string")

            if host and database:
                identifier = f"{host}_{database}"
            elif host:
                identifier = f"{host}"
            elif jdbc_string:
                identifier = f"{jdbc_string}"
            else:
                print(f"Potentially broken config: host: {host}, database: {database}, jdbc: {jdbc_string}")
            
            if grouped_endpoints.get(identifier):
                grouped_endpoints[identifier].append(config_props)
            else:
                grouped_endpoints[identifier] = [config_props]
        
        for group in track(grouped_endpoints):
            try:
                ext = EF2SqlExtension(dt, group, grouped_endpoints[group], credential_vault_id, pre_cron)
                extension_file = Path(directory, f"{ext.name.replace(':', '_')}", "src", "extension.yaml")
                extension_file.parent.mkdir(parents=True, exist_ok=True)
                extension_file.write_text(yaml.dump(ext.dictionary, sort_keys=False, Dumper=MyDumper, default_flow_style=False))
                activation_file = Path(directory, f"{ext.name.replace(':', '_')}", f"activation.json")
                activation_file.write_text(json.dumps(ext.activation_config))
                extension_zip = Path(extension_file.parent, dtcli.constants.EXTENSION_ZIP)
                dtcli.building.build_and_sign(extension_dir_path=extension_file.parent, extension_zip_path=extension_zip, 
                                            extension_zip_sig_path=Path(extension_file.parent.parent, dtcli.constants.EXTENSION_ZIP_SIG).absolute(), certificate_file_path=cert_file_path, private_key_file_path=private_key_path,
                                            target_dir_path=extension_file.parent.parent)
                
                built_extension_zip = Path(extension_file.parent.parent, f"{ext.name.replace(':', '_')}-{ext.version}.zip".replace(":", "_"))

                dtcli.server_api.validate(extension_zip_file=built_extension_zip, tenant_url=dt_url, api_token=dt_token)

                if upload:
                    dtcli.server_api.upload(extension_zip_file=built_extension_zip, tenant_url=dt_url, api_token=dt_token)
                    dt.extensions_v2.put_environment_config(extension_name=ext.name, extension_version=ext.version)
                if create_config:
                    dt.extensions_v2.post_monitoring_configurations(ext.name, [MonitoringConfigurationDto(scope="ag_group-default", configuration=ext.activation_config)])
            except Exception as e:
                print(f"Error handling group {group}: {e}")
            
        return
    
    else:
        for index, row in track(df.iterrows()):
            try:
                config_props = json.loads(row["properties"])
                ext = EF2SqlExtension(dt, row["endpointName"], config_props, credential_vault_id, pre_cron)
                extension_file = Path(directory, f"{ext.name.replace(':', '_')}", "src", "extension.yaml")
                extension_file.parent.mkdir(parents=True, exist_ok=True)
                extension_file.write_text(yaml.safe_dump(ext.dictionary, sort_keys=False))
                activation_file = Path(directory, f"{ext.name.replace(':', '_')}", f"activation.json")
                activation_file.write_text(json.dumps(ext.activation_config))
                extension_zip = Path(extension_file.parent, dtcli.constants.EXTENSION_ZIP)
                dtcli.building.build_and_sign(extension_dir_path=extension_file.parent, extension_zip_path=extension_zip, 
                                            extension_zip_sig_path=Path(extension_file.parent.parent, dtcli.constants.EXTENSION_ZIP_SIG).absolute(), certificate_file_path=cert_file_path, private_key_file_path=private_key_path,
                                            target_dir_path=extension_file.parent.parent)
                built_extension_zip = Path(extension_file.parent, f"{ext.name.replace(':', '_')}-{ext.version}.zip".replace(":", "_"))
                
                dtcli.server_api.validate(extension_zip_file=built_extension_zip, tenant_url=dt_url, api_token=dt_token)

                if upload:
                    dtcli.server_api.upload(extension_zip_file=built_extension_zip, tenant_url=dt_url, api_token=dt_token)
                    dt.extensions_v2.put_environment_config(extension_name=ext.name, extension_version=ext.version)
                if create_config:
                    dt.extensions_v2.post_monitoring_configurations(ext.name, [MonitoringConfigurationDto(scope="ag_group-default", configuration=ext.activation_config)])
            except Exception as e:
                print(f"Error handling endpoint {row['endpointName']}: {e}")


if __name__ == "__main__":
    app()



from dynatrace import Dynatrace
from dynatrace.http_client import TOO_MANY_REQUESTS_WAIT
from rich.progress import track
from rich import print

from typing import Optional, List, Dict, Union
from urllib.parse import urlparse

from .models import CustomDBQuery
from .utils import slugify

TIMEFRAME = "now-1y"


def lookup_columns_for_query(dt: Dynatrace, query_name: str):
    column_names = []
    query = f'custom.db.query:filter(eq(query_name,"{query_name}")):splitBy(column,query_name)'
    metric_series_collections = dt.metrics.query(
        query, time_from=TIMEFRAME, resolution="Inf"
    )
    for collection in metric_series_collections:
        for series in collection.data:
            column = series.dimension_map.get("column")
            if column:
                column_names.append(column)

    return column_names


def queries_from_ef1_config(properties: dict):
    properties.update(
        {
            "database_type": properties["database_type"],
            "group_name": properties["group_name"],
            "database_host": properties["database_host"],
            "database_name": properties["database_name"],
            "database_username": properties["database_username"],
            "custom_device_name": properties["custom_device_name"],
        }
    )

    configured_queries: List[CustomDBQuery] = []
    query_index = 1
    while query_index < 11:
        if properties[f"query_{query_index}_value"]:
            configured_queries.append(
                CustomDBQuery(
                    properties[f"query_{query_index}_name"],
                    properties[f"query_{query_index}_schedule"],
                    properties[f"query_{query_index}_value"],
                    properties[f"query_{query_index}_value_columns"],
                    properties[f"query_{query_index}_dimension_columns"],
                    properties[f"query_{query_index}_extra_dimensions"],
                )
            )
        query_index += 1
    return configured_queries


def ef2_datasource(db_type: str):
    if db_type == "Oracle":
        return "sqlOracle"
    elif db_type == "DB2":
        return "sqlDb2"
    elif db_type == "SQL Server":
        return "sqlServer"
    elif db_type == "MySQL":
        return "sqlMySql"
    elif db_type == "PostgreSQL":
        return "sqlPostgres"
    elif db_type == "SAP HANA":
        return "sqlHana"
    else:
        raise Exception(f"Unsupported database type: {db_type}")


def activation(endpoint_name: str, props: dict, credential_vault_id: str = None):

    database_type = props["database_type"]
    host = props.get("database_host") or None
    port = int(props.get("database_port")) or None
    database_name = props.get("database_name") or None
    database_username = props.get("database_username") or None
    oracle_listener_type = props.get("oracle_listener_type") or None
    jdbc_connection_string = props.get("jdbc_connection_string") or None

    endpoint = {}

    if jdbc_connection_string:
        if database_type not in ["Oracle, PostgreSQL", "SQL Server"]:
            raise Exception(f"{database_type} datasource does not currently support connection strings.")
        endpoint['useConnectionString'] = True
        endpoint['connectionString'] = jdbc_connection_string
        # try:
        #     # try parsing out some details from the connection string
        #     print(jdbc_connection_string)
        #     r = urlparse(jdbc_connection_string)
        #     print(r.scheme)
        # except Exception as e:
        #     print(f"Can't parse jdbc '{jdbc_connection_string}': {e}")
        endpoint['host'] = "SET-ME"
        endpoint['port'] = 1234
        endpoint['databaseName'] = "SET-ME"
    else:
        endpoint['host'] = host
        endpoint['port'] = port

    if credential_vault_id:
        endpoint['authentication'] = {
            "scheme": "basic",
            "useCredentialVault": True,
            "credentialVaultId": credential_vault_id
        }
    else:
        endpoint['authentication'] = {
            "scheme": "basic",
            "useCredentialVault": False,
            "username": database_username,
            "password": "changeme"
        }

    if database_type == "Oracle":
        endpoint['databaseIdentifier'] = "sid" if oracle_listener_type == "SID" else "serviceName"
        endpoint.update({"SID" if oracle_listener_type == "SID" else "serviceName": database_name})
    else:
        endpoint['databaseName'] = database_name

    activation_config = {
        "enabled": False,
        "description": f"Migrated extension from EF1 db queries config {endpoint_name}.",
        "version": "1.0.0",
        f"{ef2_datasource(props['database_type'])}Remote": {
            "endpoints": [endpoint]
        }
    }

    if database_type == "Oracle":
        activation_config[f"{ef2_datasource(props['database_type'])}Remote"].update({"licenseAccepted": True})

    return activation_config


class EF2SqlExtension:
    def __init__(
        self, dt: Dynatrace, endpoint_name: str, ef1_config_properties: Union[Dict, List], credential_vault_id = None, pre_cron = False
    ) -> None:
        

        if type(ef1_config_properties) == list:
            # have to merge queries from multiple configs into one
            queries = []
            db_type = ef2_datasource(ef1_config_properties[0]["database_type"])
            for conf_prop in ef1_config_properties:
                queries.extend(queries_from_ef1_config(conf_prop))
            self.activation_config = activation(endpoint_name, ef1_config_properties[0], credential_vault_id)
        else:
            db_type = ef2_datasource(ef1_config_properties["database_type"])
            queries = queries_from_ef1_config(ef1_config_properties)
            self.activation_config = activation(endpoint_name, ef1_config_properties, credential_vault_id)

        
        extension = {
            "name": f"custom:db.query.{slugify(endpoint_name)[:30] if len(slugify(endpoint_name)) > 30 else slugify(endpoint_name)}",
            "version": "1.0.0",
            "minDynatraceVersion": "1.301" if not pre_cron else "1.299",
            "author": {"name": "Dynatrace"},
            db_type: [],
        }

        group_number = 1
        subgroup_number = 0

        group = {
                "group": f"queries-{group_number}",
                "subgroups": []
            }

        for query in queries:
            if subgroup_number == 10:
                extension[db_type].append(group)
                group_number+=1
                group = {
                    "group": f"queries-{group_number}",
                    "subgroups": []
                }
            metric_columns = query.value_columns
            if not metric_columns:
                metric_columns = lookup_columns_for_query(dt, query.name)
            else:
                metric_columns = metric_columns

            safe_query_name = slugify(query.name)

            subgroup = {
                "subgroup": query.name,
                "query": query.value,
                "metrics": [],
                "dimensions": [{"key": "query_name", "value": f"const:{query.name}"}],
            }

            if not pre_cron:
                subgroup.update({"schedule": query.schedule})

            for column in metric_columns:
                subgroup["metrics"].append(
                    {
                        "key": f"{safe_query_name}.{slugify(column)}",
                        "value": f"col:{column}",
                        "type": "gauge",
                    }
                )

            for column in query.dimension_columns:
                subgroup['dimensions'].append(
                    {
                        "key": column.lower() ,
                        "value": f"col:{column}"
                    }
                )

            for dimension_pair in query.extra_dimensions:
                key, value = dimension_pair.split("=")
                subgroup['dimensions'].append(
                    {
                        "key": key.lower(),
                        "value": f"const:{value}"
                    }
                )

            subgroup_number += 1
            group['subgroups'].append(subgroup)


        extension[db_type].append(group)

        self.name = "custom:" + extension["name"].split(":")[1]
        self.version = extension["version"]
        self.dictionary = extension



from typing import List, Dict, Optional

class CustomDBQuery:
    def __init__(
        self,
        name: str,
        schedule: str,
        value: str,
        value_columns: Optional[str] = None,
        dimension_columns: Optional[str] = None,
        extra_dimensions: Optional[str] = None
    ) -> None:
        self.name = f"{name}"
        self.schedule = schedule if schedule else "* * * * *"
        self.value = value.replace("\n", " ")
        self.value_columns = value_columns.split(",") if value_columns else []
        self.dimension_columns = dimension_columns.split(",") if dimension_columns else []
        self.extra_dimensions = extra_dimensions.split(",") if extra_dimensions else []
